import pytest

from cyberpandas import dtypes


@pytest.mark.parametrize('value', [
    "192.168.0.1",
    3232235521,
])
@pytest.mark.xfail
def test_is_ipv4(value):
    assert dtypes.is_ipv4(value)


@pytest.mark.parametrize('value', [
    '123.123.1',
    2 ** 32,
])
def test_is_not_ipv4(value):
    assert not dtypes.is_ipv4(value)



import ipaddress

import pytest

from cyberpandas import parser, IPArray


@pytest.mark.parametrize('values', [
    [u'192.168.1.1',
     u'2001:0db8:85a3:0000:0000:8a2e:0370:7334'],
    [3232235777,
     42540766452641154071740215577757643572],
    [b'\xc0\xa8\x01\x01',
     b' \x01\r\xb8\x85\xa3\x00\x00\x00\x00\x8a.\x03ps4'],
])
def test_to_ipaddress(values):
    result = parser.to_ipaddress(values)
    expected = IPArray.from_pyints([
        3232235777,
        42540766452641154071740215577757643572
    ])
    assert result.equals(expected)


def test_to_ipaddress_edge():
    ip_int = 2 ** 64
    result = parser.to_ipaddress([ip_int]).to_pyipaddress()[0]
    assert int(result) == ip_int


def test_to_ipaddress_scalar():
    result = parser.to_ipaddress(1)
    expected = parser.to_ipaddress([1])
    assert len(result) == 1
    assert all(result == expected)


@pytest.mark.parametrize('val, expected', [
    (u'192.168.1.1', ipaddress.IPv4Address(u'192.168.1.1')),
    (100, ipaddress.IPv4Address(100)),
    (ipaddress.IPv4Address(100), ipaddress.IPv4Address(100)),
    (2**64, ipaddress.IPv6Address(2**64)),
    (u'192.168.0.0/28', ipaddress.IPv4Network(u'192.168.0.0/28')),
    (ipaddress.IPv4Network(u'192.168.0.0/28'),
     ipaddress.IPv4Network(u'192.168.0.0/28')),
    (u'2001:db00::0/24', ipaddress.IPv6Network(u'2001:db00::0/24')),
])
def test_as_ip_object(val, expected):
    result = parser._as_ip_object(val)
    assert result == expected


@pytest.mark.parametrize("val", [
    u"129", -1
])
def test_as_ip_object_raises(val):
    with pytest.raises(ValueError):
        parser._as_ip_object(val)



import ipaddress
import operator

import pytest
import six
from hypothesis.strategies import integers, lists, tuples
from hypothesis import given, example

import numpy as np
import numpy.testing as npt
import pandas as pd
import cyberpandas as ip
import pandas.util.testing as tm
from cyberpandas.common import _U8_MAX


def test_make_container():
    values = ip.IPArray.from_pyints([1, 2, 3])
    npt.assert_array_equal(
        values.data,
        np.array([(0, 1),
                  (0, 2),
                  (0, 3)], dtype=values.dtype._record_type)
    )


def test_repr_works():
    values = ip.IPArray.from_pyints([0, 1, 2, 3, 2**32, 2**64 + 1])
    result = repr(values)
    if six.PY2:
        expected = ("IPArray([u'0.0.0.0', u'0.0.0.1', u'0.0.0.2', u'0.0.0.3', "
                    "u'::1:0:0', u'::1:0:0:0:1'])")
    else:
        expected = ("IPArray(['0.0.0.0', '0.0.0.1', '0.0.0.2', '0.0.0.3', "
                    "'::1:0:0', '::1:0:0:0:1'])")
    assert result == expected


def test_isna():
    v = ip.IPArray.from_pyints([0, 2, 2 ** 64, 2 ** 64 + 1, 2 ** 64 + 2])
    r1 = v.isna()
    r2 = pd.isna(v)
    expected = np.array([True, False, False, False, False])

    np.testing.assert_array_equal(r1, expected)
    np.testing.assert_array_equal(r2, expected)


def test_array():
    v = ip.IPArray.from_pyints([1, 2, 3])
    result = np.array(v)
    expected = np.array([
        ipaddress.IPv4Address(1),
        ipaddress.IPv4Address(2),
        ipaddress.IPv4Address(3),
    ])
    tm.assert_numpy_array_equal(result, expected)


def test_tolist():
    v = ip.IPArray.from_pyints([1, 2, 3])
    result = v.tolist()
    expected = [(0, 1), (0, 2), (0, 3)]
    assert result == expected


def test_to_pyipaddress():
    v = ip.IPArray.from_pyints([1, 2, 3])
    result = v.to_pyipaddress()
    expected = [
        ipaddress.ip_address(1),
        ipaddress.ip_address(2),
        ipaddress.ip_address(3),
    ]
    assert result == expected


def test_isip():
    v = ip.to_ipaddress([
        u'192.168.1.1',
        u'2001:0db8:85a3:0000:0000:8a2e:0370:7334',
    ])
    result = v.is_ipv4
    expected = np.array([True, False])
    tm.assert_numpy_array_equal(result, expected)

    result = v.is_ipv6
    expected = np.array([False, True])
    tm.assert_numpy_array_equal(result, expected)


def test_equality():
    v1 = ip.to_ipaddress([
        u'192.168.1.1',
        u'2001:0db8:85a3:0000:0000:8a2e:0370:7334',
    ])
    assert np.all(v1 == v1)
    assert v1.equals(v1)

    v2 = ip.to_ipaddress([
        u'192.168.1.2',
        u'2001:0db8:85a3:0000:0000:8a2e:0370:7334',
    ])
    result = v1 == v2
    expected = np.array([False, True])
    tm.assert_numpy_array_equal(result, expected)

    result = bool(v1.equals(v2))
    assert result is False

    with pytest.raises(TypeError):
        v1.equals("a")


@pytest.mark.parametrize('op', [
    operator.lt,
    operator.le,
    operator.ge,
    operator.gt,
])
@pytest.mark.skipif(six.PY2, reason="Flexible comparisons")
def test_comparison_raises(op):
    arr = ip.IPArray([0, 1, 2])
    with pytest.raises(TypeError):
        op(arr, 'a')

    with pytest.raises(TypeError):
        op('a', arr)


@given(
    tuples(
        lists(integers(min_value=0, max_value=2**128 - 1)),
        lists(integers(min_value=0, max_value=2**128 - 1))
    ).filter(lambda x: len(x[0]) == len(x[1]))
)
@example((1, 1))
@example((0, 0))
@example((0, 1))
@example((1, 0))
@example((1, 2))
@example((2, 1))
@pytest.mark.skip(reason="Flaky")
def test_ops(tup):
    a, b = tup
    v1 = ip.IPArray(a)
    v2 = ip.IPArray(b)

    r1 = v1 <= v2
    r2 = v2 >= v1
    tm.assert_numpy_array_equal(r1, r2)


@pytest.mark.xfail(reason='upstream')
def test_value_counts():
    x = ip.IPArray([0, 0, 1])
    result = x.value_counts()
    assert len(result)


def test_iter_works():
    x = ip.IPArray([0, 1, 2])
    result = list(x)
    expected = [
        ipaddress.IPv4Address(0),
        ipaddress.IPv4Address(1),
        ipaddress.IPv4Address(2),
    ]
    assert result == expected


def test_topyints():
    values = [0, 1, 2**32]
    arr = ip.IPArray(values)
    result = arr.to_pyints()
    assert result == values


@pytest.mark.parametrize('prop', [
    'version',
    'is_multicast',
    'is_private',
    'is_global',
    'is_unspecified',
    'is_reserved',
    'is_loopback',
    'is_link_local',
])
def test_attributes(prop):
    addrs = [ipaddress.ip_address(0),
             ipaddress.ip_address(1)]
    arr = ip.IPArray(addrs)
    result = getattr(arr, prop)
    expected = np.array([getattr(addr, prop)
                         for addr in addrs])
    tm.assert_numpy_array_equal(result, expected)


def test_isin_all4():
    s = ip.IPArray([u'192.168.1.1', u'255.255.255.255'])
    result = s.isin([u'192.168.1.0/24'])
    expected = np.array([True, False])
    tm.assert_numpy_array_equal(result, expected)


def test_isin_all6():
    s = ip.IPArray([u'2001:db8::1', u'2001:db9::1'])
    result = s.isin([u'2001:db8::0/96'])
    expected = np.array([True, False])
    tm.assert_numpy_array_equal(result, expected)


def test_isin_mix():
    s = ip.IPArray([u'192.168.1.1', u'255.255.255.255',
                    u'2001:db8::1', u'2001:db9::1'])

    result = s.isin([u'192.168.1.0/24'])
    expected = np.array([True, False, False, False])
    tm.assert_numpy_array_equal(result, expected)

    result = s.isin([u'2001:db8::0/96'])
    expected = np.array([False, False, True, False])
    tm.assert_numpy_array_equal(result, expected)

    result = s.isin([u'192.168.1.0/24', u'2001:db8::0/96'])
    expected = np.array([True, False, True, False])
    tm.assert_numpy_array_equal(result, expected)

    s = ip.IPArray([u'192.168.1.1', u'192.168.1.2',
                    u'255.255.255.255'])
    result = s.isin([u'192.168.1.0/24'])
    expected = np.array([True, True, False])
    tm.assert_numpy_array_equal(result, expected)


def test_isin_iparray():
    s = ip.IPArray([10, 20, 20, 30])
    result = s.isin(ip.IPArray([30, 20]))
    expected = np.array([False, True, True, True])
    tm.assert_numpy_array_equal(result, expected)


def test_getitem_scalar():
    ser = ip.IPArray([0, 1, 2])
    result = ser[1]
    assert result == ipaddress.ip_address(1)


def test_getitem_slice():
    ser = ip.IPArray([0, 1, 2])
    result = ser[1:]
    expected = ip.IPArray([1, 2])
    assert result.equals(expected)


@pytest.mark.parametrize('value', [
    u'0.0.0.10',
    10,
    ipaddress.ip_address(10),
])
def test_setitem_scalar(value):
    ser = ip.IPArray([0, 1, 2])
    ser[1] = ipaddress.ip_address(value)
    expected = ip.IPArray([0, 10, 2])
    assert ser.equals(expected)


def test_setitem_array():
    ser = ip.IPArray([0, 1, 2])
    ser[[1, 2]] = [10, 20]
    expected = ip.IPArray([0, 10, 20])
    assert ser.equals(expected)


def test_bytes_roundtrip():
    arr = ip.IPArray([1, 2, 3, _U8_MAX + 10])
    bytestring = arr.to_bytes()
    assert isinstance(bytestring, bytes)

    result = ip.IPArray.from_bytes(bytestring)
    assert result.equals(arr)


def test_unique():
    arr = ip.IPArray([3, 3, 1, 2, 3, _U8_MAX + 1])
    result = arr.unique()
    assert isinstance(result, ip.IPArray)

    result = result.astype(object)
    expected = pd.unique(arr.astype(object))
    tm.assert_numpy_array_equal(result, expected)


def test_factorize():
    arr = ip.IPArray([3, 3, 1, 2, 3, _U8_MAX + 1])
    labels, uniques = arr.factorize()
    expected_labels, expected_uniques = pd.factorize(arr.astype(object))

    assert isinstance(uniques, ip.IPArray)

    uniques = uniques.astype(object)
    tm.assert_numpy_array_equal(labels, expected_labels)
    tm.assert_numpy_array_equal(uniques, expected_uniques)


@pytest.mark.parametrize('values', [
    [0, 1, 2],
])
def test_from_ndarray(values):
    result = ip.IPArray(np.asarray(values))
    expected = ip.IPArray(values)
    assert result.equals(expected)


@pytest.mark.parametrize('start, stop, step, expected', [
    (1, 3, None, [1, 2]),
    (u'0.0.0.1', u'0.0.0.3', None, [1, 2]),
    (2**64 + 1, 2**64 + 3, None, [2**64 + 1, 2**64 + 2]),
    (u'::1:0:0:0:1', u'::1:0:0:0:3', None, [2**64 + 1, 2**64 + 2]),
    (2**64 - 1, 2**64 + 2, None, [2**64 - 1, 2**64, 2**64 + 1]),
    (u'::ffff:ffff:ffff:ffff', u'::1:0:0:0:2', None,
     [2**64 - 1, 2**64, 2**64 + 1]),
    (1, 6, 2, [1, 3, 5]),
    (u'0.0.0.1', u'0.0.0.6', u'0.0.0.2', [1, 3, 5]),
])
def test_ip_range(start, stop, step, expected):
    result = ip.ip_range(start, stop, step)
    expected = ip.IPArray(expected)
    assert result.equals(expected)


@pytest.mark.parametrize('addresses', [
    [u'0.0.0.0', u'192.168.1.1', u'::1:1:0:0:0:1']
])
@pytest.mark.parametrize('v4_prefixlen, v6_prefixlen', [
    (32, 128),
    (24, 96),
    (16, 64),
    (8, 8),
])
@pytest.mark.parametrize('op', ['netmask', 'hostmask'])
def test_mask(op, v4_prefixlen, v6_prefixlen, addresses):
    is_v6 = [':' in x for x in addresses]
    prefixes = [v6_prefixlen if v6 else v4_prefixlen for v6 in is_v6]
    networks = [
        ipaddress.ip_network(u"{}/{}".format(addr, prefix), strict=False)
        for addr, prefix in zip(addresses, prefixes)
    ]
    expected = [getattr(net, op) for net in networks]
    call = operator.methodcaller(op, v4_prefixlen=v4_prefixlen,
                                 v6_prefixlen=v6_prefixlen)
    result = list(call(ip.IPArray(addresses)))
    if op == 'hostmask':
        # ipaddress will return an IPv6(0), which doesn't compare equal
        # to an IPv4(0), our result.
        expected = [int(x) for x in expected]
        result = [int(x) for x in result]

    assert result == expected


def test_netmask_basic():
    arr = ip.IPArray([u'192.0.0.0', u'1:1::'])
    result = arr.netmask(v4_prefixlen=16, v6_prefixlen=32)
    expected = ip.IPArray([u'255.255.0.0', u'ffff:ffff::'])
    assert result.equals(expected)

    result = pd.Series(arr, name='foo').ip.netmask(v4_prefixlen=16,
                                                   v6_prefixlen=32)
    assert result.name == 'foo'
    assert result.values.equals(expected)


def test_hostmask_basic():
    arr = ip.IPArray([u'192.0.0.0', u'1:1::'])
    result = arr.hostmask(v4_prefixlen=16, v6_prefixlen=32)
    expected = ip.IPArray([u'0.0.255.255',
                           u'::ffff:ffff:ffff:ffff:ffff:ffff'])
    assert result.equals(expected)

    result = pd.Series(arr, name='foo').ip.hostmask(v4_prefixlen=16,
                                                    v6_prefixlen=32)
    assert result.name == 'foo'
    assert result.values.equals(expected)


def test_apply_mask():
    arr = ip.IPArray([u'216.3.128.0', u'192.168.100.0', u'1::1:12'])
    mask = arr.netmask(v4_prefixlen=24, v6_prefixlen=112)
    result = arr.mask(mask)
    expected = ip.IPArray([u'216.3.128.0', u'192.168.100.0', u'1::1:0'])
    assert result.equals(expected)

    result = pd.Series(arr, name='test').ip.mask(mask)
    expected = pd.Series(expected, name='test')
    tm.assert_series_equal(result, expected)



"""Utilities for working with IP address data."""
import struct

import six


def to_bytes(n, length, byteorder='big'):
    # https://stackoverflow.com/a/20793663/1889400
    h = '%x' % n
    s = ('0' * (len(h) % 2) + h).zfill(length * 2).decode('hex')
    return s if byteorder == 'big' else s[::-1]


def pack(ip):
    # type: (int) -> bytes
    if six.PY2:
        return to_bytes(ip, length=16, byteorder='big')
    else:
        return ip.to_bytes(16, byteorder='big')


def unpack(ip):
    # type: (T.Tuple[int, int]) -> int, int
    # Recipe 3.5 from Python Cookbook 3rd ed. (p. 90)
    # int.from_bytes(data, 'big') for Py3+
    hi, lo = struct.unpack(">QQ", ip)
    return hi, lo


def combine(hi, lo):
    # type: (int, int) -> int
    """Combine the hi and lo bytes into the final ip address."""
    return (hi << 64) + lo



"""Tests involving pandas, not just the new array.
"""
import ipaddress

import pytest
import numpy as np
from hypothesis.strategies import integers, lists
from hypothesis import given
import pandas as pd
from pandas.core.internals import ExtensionBlock
import pandas.util.testing as tm

import cyberpandas as ip


# ----------------------------------------------------------------------------
# Block Methods
# ----------------------------------------------------------------------------


def test_concatenate_blocks():
    v1 = ip.IPArray.from_pyints([1, 2, 3])
    s = pd.Series(v1, index=pd.RangeIndex(3), fastpath=True)
    result = pd.concat([s, s], ignore_index=True)
    expected = pd.Series(ip.IPArray.from_pyints([1, 2, 3, 1, 2, 3]))
    tm.assert_series_equal(result, expected)


# ----------------------------------------------------------------------------
# Public Constructors
# ----------------------------------------------------------------------------


def test_series_constructor():
    v = ip.IPArray.from_pyints([1, 2, 3])
    result = pd.Series(v)
    assert result.dtype == v.dtype
    assert isinstance(result._data.blocks[0], ExtensionBlock)


def test_dataframe_constructor():
    v = ip.IPArray.from_pyints([1, 2, 3])
    df = pd.DataFrame({"A": v})
    assert isinstance(df.dtypes['A'], ip.IPType)
    assert df.shape == (3, 1)
    str(df)


def test_dataframe_from_series_no_dict():
    s = pd.Series(ip.IPArray([1, 2, 3]))
    result = pd.DataFrame(s)
    expected = pd.DataFrame({0: s})
    tm.assert_frame_equal(result, expected)

    s = pd.Series(ip.IPArray([1, 2, 3]), name='A')
    result = pd.DataFrame(s)
    expected = pd.DataFrame({'A': s})
    tm.assert_frame_equal(result, expected)


def test_dataframe_from_series():
    s = pd.Series(ip.IPArray([0, 1, 2]))
    c = pd.Series(pd.Categorical(['a', 'b']))
    result = pd.DataFrame({"A": s, 'B': c})
    assert isinstance(result.dtypes['A'], ip.IPType)


def test_getitem_scalar():
    ser = pd.Series(ip.IPArray([0, 1, 2]))
    result = ser[1]
    assert result == ipaddress.ip_address(1)


def test_getitem_slice():
    ser = pd.Series(ip.IPArray([0, 1, 2]))
    result = ser[1:]
    expected = pd.Series(ip.IPArray([1, 2]), index=range(1, 3))
    tm.assert_series_equal(result, expected)


def test_setitem_scalar():
    ser = pd.Series(ip.IPArray([0, 1, 2]))
    ser[1] = ipaddress.ip_address(10)
    expected = pd.Series(ip.IPArray([0, 10, 2]))
    tm.assert_series_equal(ser, expected)


# --------------
# Public Methods
# --------------


@given(lists(integers(min_value=1, max_value=2**128 - 1)))
def test_argsort(ints):
    pass
    # result = pd.Series(ip.IPArray(ints)).argsort()
    # expected = pd.Series(ints).argsort()
    # tm.assert_series_equal(result.ip.to_pyints(), expected)


# --------
# Accessor
# --------

def test_non_ip_raises():
    s = pd.Series([1, 2])

    with pytest.raises(AttributeError) as m:
        s.ip.is_ipv4

    assert m.match("Cannot use 'ip' accessor on objects of dtype 'int.*")


def test_accessor_works():
    s = pd.Series(ip.IPArray([0, 1, 2, 3]))
    s.ip.is_ipv4


def test_accessor_frame():
    s = pd.DataFrame({"A": ip.IPArray([0, 1, 2, 3])})
    s['A'].ip.is_ipv4


# ---------
# Factorize
# ---------


@pytest.mark.xfail(reason="TODO")
def test_factorize():
    arr = ip.IPArray([1, 1, 10, 10])
    labels, uniques = pd.factorize(arr)

    expected_labels = np.array([0, 0, 1, 1])
    tm.assert_numpy_array_equal(labels, expected_labels)

    expected_uniques = ip.IPArray([1, 10])
    assert uniques.equals(expected_uniques)


@pytest.mark.xfail(reason="TODO")
def test_groupby_make_grouper():
    df = pd.DataFrame({"A": [1, 1, 2, 2],
                       "B": ip.IPArray([1, 1, 2, 2])})
    gr = df.groupby("B")
    result = gr.grouper.groupings[0].grouper
    assert result.equals(df.B.values)


@pytest.mark.xfail(reason="TODO")
def test_groupby_make_grouper_groupings():
    df = pd.DataFrame({"A": [1, 1, 2, 2],
                       "B": ip.IPArray([1, 1, 2, 2])})
    p1 = df.groupby("A").grouper.groupings[0]
    p2 = df.groupby("B").grouper.groupings[0]

    result = {int(k): v for k, v in p2.groups.items()}
    assert result.keys() == p1.groups.keys()
    for k in result.keys():
        assert result[k].equals(p1.groups[k])



import ipaddress

import numpy as np
import six

from .ip_array import IPArray
from .common import _U8_MAX


def _as_int(ip):
    # type: (Union[int, str, IPv4Address, IPv6Address]) -> int
    if isinstance(ip, six.string_types):
        ip = ipaddress.ip_address(ip)
    return int(ip)


def _crosses_boundary(lo, hi):
    return (lo <= _U8_MAX) == (hi <= _U8_MAX)


def ip_range(start=None, stop=None, step=None):
    """Generate a range of IP Addresses

    Parameters
    ----------
    start : int, str, IPv4Address, or IPv6Address, optional
        Start of interval.  The interval includes this value.  The default
        start value is 0.
    start : int, str, IPv4Address, or IPv6Address, optional
        End of interval.  The interval does not include this value.
    step : int, optional
        Spacing between values.  For any output `out`, this is the distance
        between two adjacent values, ``out[i+1] - out[i]``.  The default
        step size is 1.  If `step` is specified as a position argument,
        `start` must also be given.

    Returns
    -------
    IPArray

    Notes
    -----
    Performance will worsen if either of `start` or `stop` are larger than
    2**64.

    Examples
    --------
    From integers

    >>> ip_range(1, 5)
    IPArray(['0.0.0.1', '0.0.0.2', '0.0.0.3', '0.0.0.4'])

    Or strings

    >>> ip_range('0.0.0.1', '0.0.0.5')
    IPArray(['0.0.0.1', '0.0.0.2', '0.0.0.3', '0.0.0.4'])

    Or `ipaddress` objects

    >>> ip_range(ipaddress.IPv4Address(1), ipaddress.IPv4Address(5))
    IPArray(['0.0.0.1', '0.0.0.2', '0.0.0.3', '0.0.0.4'])
    """
    if start is not None:
        start = _as_int(start)
    if stop is not None:
        stop = _as_int(stop)
    if step is not None:
        step = _as_int(step)
    arr = IPArray(np.arange(start, stop, step))
    return arr
